import asyncio
import time
from datetime import datetime, timedelta
from typing import Any, AsyncGenerator, Dict, Sequence, Union


def to_ms_timestamp(value: Union[None, int, float, timedelta, datetime]):
    """
    covert to timestamp
    :param value:
    :return:
    """
    if isinstance(value, datetime):
        return round(value.timestamp() * 1000)
    if isinstance(value, timedelta):
        value = value.total_seconds()
    value = value or 0
    return round((time.time() + value) * 1000)


def timestamp_ms_now() -> int:
    """
    now timestamp
    :return:
    """
    return round(time.time() * 1000)


def ms_to_datetime(ms: int) -> datetime:
    """
    ms to datetime
    :param ms:
    :return:
    """
    return datetime.fromtimestamp(ms / 1000)


async def poll(step: float = 0.5) -> AsyncGenerator[float, None]:
    """
    make a async generator
    :param step:
    :return:
    """
    loop = asyncio.get_event_loop()
    start = loop.time()
    while True:
        before = loop.time()
        yield before - start
        after = loop.time()
        wait = max([0, step - after + before])
        await asyncio.sleep(wait)


def args_to_string(args: Sequence[Any], kwargs: Dict[str, Any]) -> str:
    arguments = ""
    if args:
        arguments = ", ".join(map(repr, args))
    if kwargs:
        if arguments:
            arguments += ", "
        arguments += ", ".join(f"{k}={v!r}" for k, v in sorted(kwargs.items()))
    return arguments



import logging
import sys


def init_logging(verbose):
    logger = logging.getLogger("rearq")
    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    sh = logging.StreamHandler(sys.stdout)
    sh.setLevel(logging.DEBUG)
    sh.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s - %(name)s:%(lineno)d - %(levelname)s - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    )
    logger.addHandler(sh)



import datetime
import logging
from typing import Any, Callable, Dict, Optional, Tuple, Union
from uuid import uuid4

from aioredis.commands import MultiExec
from crontab import CronTab

from rearq.constants import delay_queue, job_key_prefix, result_key_prefix
from rearq.job import Job, JobDef
from rearq.utils import timestamp_ms_now, to_ms_timestamp

logger = logging.getLogger("rearq.jobs")


class Task:
    expires_extra_ms = 86_400_000
    job_def: JobDef

    def __init__(self, bind: bool, function: Callable, queue: str, rearq, job_retry: int):

        self.job_retry = job_retry
        self.queue = queue
        self.rearq = rearq
        self.function = function
        self.bind = bind

    async def delay(
        self,
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        job_id: str = None,
        countdown: Union[float, datetime.timedelta] = 0,
        eta: Optional[datetime.datetime] = None,
        expires: Optional[Union[float, datetime.datetime]] = None,
        job_retry: int = 0,
    ):
        if not job_id:
            job_id = uuid4().hex
        if countdown:
            defer_ts = to_ms_timestamp(countdown)
        elif eta:
            defer_ts = to_ms_timestamp(eta)
        else:
            defer_ts = timestamp_ms_now()
        enqueue_ms = timestamp_ms_now()
        expires_ms = (
            to_ms_timestamp(expires) if expires else defer_ts - enqueue_ms + self.expires_extra_ms
        )
        job_key = job_key_prefix + job_id
        redis = self.rearq.get_redis()
        pipe = redis.pipeline()
        pipe.exists(job_key)
        pipe.exists(result_key_prefix + job_id)
        job_exists, job_result_exists = await pipe.execute()
        if job_exists or job_result_exists:
            logger.warning(
                f"Job {job_id} exists, job_exists={job_exists}, job_result_exists={job_result_exists}"
            )
            return Job(redis, job_id, self.queue,)

        p = redis.pipeline()  # type:MultiExec
        p.psetex(
            job_key,
            expires_ms,
            JobDef(
                function=self.function.__name__,
                args=args,
                kwargs=kwargs,
                job_retry=job_retry or self.job_retry,
                enqueue_ms=enqueue_ms,
                queue=self.queue,
                job_id=job_id,
            ).json(),
        )

        if not eta and not countdown:
            p.xadd(self.queue, {"job_id": job_id})
        else:
            p.zadd(delay_queue, defer_ts, job_id)

        await p.execute()

        return Job(redis, job_id, self.queue,)


async def check_pending_msgs(
    self: Task, queue: str, group_name: str, consumer_name: str, timeout: int
):
    """
    check pending messages
    :return:
    """
    rearq = self.rearq
    redis = rearq.get_redis()
    pending_msgs = await redis.xpending(self.queue, group_name, "-", "+", 10)
    p = redis.pipeline()
    execute = False
    for msg in pending_msgs:
        msg_id, _, idle_time, times = msg
        if int(idle_time / 1000) > timeout * 2:
            execute = True
            p.xclaim(queue, group_name, consumer_name, min_idle_time=1000, id=msg_id)
    if execute:
        return await p.execute()


class CronTask(Task):
    _cron_tasks: Dict[str, "CronTask"] = {}
    next_run: int

    def __init__(
        self, bind: bool, function: Callable, queue: str, rearq, job_retry: int, cron: str
    ):
        super().__init__(bind, function, queue, rearq, job_retry)
        self.cron = CronTab(cron)
        self.set_next()

    def set_next(self):
        self.next_run = to_ms_timestamp(self.cron.next(default_utc=False))

    @classmethod
    def add_cron_task(cls, function: str, cron_task: "CronTask"):
        cls._cron_tasks[function] = cron_task

    @classmethod
    def get_cron_tasks(cls):
        return cls._cron_tasks

    def __repr__(self) -> str:
        return f"<rearq CronTask {self.function}>"



import asyncio
import functools
import logging
from functools import wraps
from ssl import SSLContext
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import aioredis
from aioredis import Redis

from rearq.constants import default_queue, delay_queue, queue_key_prefix
from rearq.exceptions import ConfigurationError, UsageError
from rearq.task import CronTask, Task, check_pending_msgs

Serializer = Callable[[Dict[str, Any]], bytes]
Deserializer = Callable[[bytes], Dict[str, Any]]

logger = logging.getLogger("rearq")


class ReArq:
    _redis: Optional[Redis] = None
    _task_map: Dict[str, Task] = {}
    _queue_task_map: Dict[str, List[str]] = {}
    _on_startup: Set[Callable] = set()
    _on_shutdown: Set[Callable] = set()

    def __init__(
        self,
        redis_host: Union[str, List[Tuple[str, int]]] = "127.0.0.1",
        redis_port: int = 6379,
        redis_password: Optional[str] = None,
        redis_db=0,
        ssl: Union[bool, None, SSLContext] = None,
        sentinel: bool = False,
        sentinel_master: str = "master",
        job_retry: int = 3,
        max_jobs: int = 10,
        keep_result_seconds: int = 3600,
        job_timeout: int = 300,
    ):
        self.job_timeout = job_timeout
        self.keep_result_seconds = keep_result_seconds
        self.max_jobs = max_jobs
        self.job_retry = job_retry

        self.sentinel = sentinel
        self.ssl = ssl
        self.sentinel_master = sentinel_master
        self.redis_db = redis_db
        self.redis_port = redis_port
        self.redis_password = redis_password
        self.redis_host = redis_host

    async def init(self):
        if self._redis:
            return

        if type(self.redis_host) is str and self.sentinel:
            raise ConfigurationError(
                "str provided for `redis_host` but `sentinel` is true, list of sentinels expected"
            )

        if self.sentinel:
            addr = self.redis_host

            async def pool_factory(*args: Any, **kwargs: Any) -> Redis:
                client = await aioredis.sentinel.create_sentinel_pool(*args, ssl=self.ssl, **kwargs)
                return client.master_for(self.sentinel_master)

        else:
            pool_factory = functools.partial(aioredis.create_pool, ssl=self.ssl)
            addr = self.redis_host, self.redis_port
        pool = await pool_factory(
            addr, db=self.redis_db, password=self.redis_password, encoding="utf8"
        )
        self._redis = Redis(pool)

    def get_redis(self) -> Redis:
        if not self._redis:
            raise UsageError("You must call .init() first!")
        return self._redis

    def get_task_map(self) -> Dict[str, Task]:
        return self._task_map

    def get_queue_tasks(self, queue: str) -> List[str]:
        tasks = self._queue_task_map.get(queue)
        return tasks

    def create_task(
        self,
        bind: bool,
        func: Callable,
        queue: Optional[str] = None,
        cron: Optional[str] = None,
        name: Optional[str] = None,
    ):

        if not callable(func):
            raise UsageError("Task must be Callable!")

        function = name or func.__name__
        if function in self._queue_task_map:
            raise UsageError("Task name must be unique!")

        if function != check_pending_msgs.__name__:
            self._queue_task_map.setdefault(queue, []).append(function)

        defaults = dict(
            function=func,
            queue=queue_key_prefix + queue if queue else default_queue,
            rearq=self,
            job_retry=self.job_retry,
            bind=bind,
        )
        if cron:
            t = CronTask(**defaults, cron=cron)
            CronTask.add_cron_task(function, t)
        else:
            t = Task(**defaults)
        self._task_map[function] = t
        return t

    def task(
        self,
        bind: bool = True,
        queue: Optional[str] = None,
        cron: Optional[str] = None,
        name: Optional[str] = None,
    ):
        def wrapper(func: Callable):
            return self.create_task(bind, func, queue, cron, name)

        return wrapper

    def on_startup(self, func: Callable):
        self._on_startup.add(func)

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    def on_shutdown(self, func: Callable):
        self._on_shutdown.add(func)

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    async def shutdown(self):
        tasks = []
        for fun in self._on_shutdown:
            tasks.append(fun())
        if tasks:
            await asyncio.gather(*tasks)

    async def startup(self):
        tasks = []
        for fun in self._on_startup:
            tasks.append(fun())
        if tasks:
            await asyncio.gather(*tasks)

    async def close(self):
        if not self._redis:
            return
        self._redis.close()
        await self._redis.wait_closed()
        self._redis = None

    async def cancel(self, job_id: str):
        """
        cancel delay task
        :param job_id:
        :return:
        """
        return await self._redis.zrem(delay_queue, job_id)



import asyncio
import importlib
import sys
from functools import wraps

import click
import uvicorn
from click import BadArgumentUsage, Context

from rearq.api.app import app
from rearq.log import init_logging
from rearq.version import VERSION
from rearq.worker import TimerWorker, Worker


def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(f(*args, **kwargs))

    return wrapper


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(VERSION, "-v", "--version")
@click.option("--verbose", default=False, is_flag=True, help="Enable verbose output.")
@click.argument("rearq", required=True)
@click.pass_context
@coro
async def cli(ctx: Context, rearq: str, verbose):
    init_logging(verbose)
    splits = rearq.split(":")
    rearq_path = splits[0]
    rearq = splits[1]
    try:
        module = importlib.import_module(rearq_path)
        rearq = getattr(module, rearq, None)

        ctx.ensure_object(dict)
        ctx.obj["rearq"] = rearq
        ctx.obj["verbose"] = verbose

    except (ModuleNotFoundError, AttributeError) as e:
        raise BadArgumentUsage(ctx=ctx, message=f"Init rearq error, {e}.")


@cli.command(help="Start rearq worker.")
@click.option("-q", "--queue", required=False, help="Queue to consume.")
@click.option("-t", "--timer", default=False, is_flag=True, help="Start a timer worker.")
@click.pass_context
@coro
async def worker(ctx: Context, queue: str, timer: bool):
    rearq = ctx.obj["rearq"]
    await rearq.init()
    if timer:
        w = TimerWorker(rearq, queue=queue, )
    else:
        w = Worker(rearq, queue=queue)
    await w.async_run()


@cli.command(help="Start rest api server.")
@click.option("--host", default="0.0.0.0", show_default=True, help="Listen host.")
@click.option("-p", "--port", default=8080, show_default=True, help="Listen port.")
@click.pass_context
def server(ctx: Context, host: str, port: int):
    rearq = ctx.obj["rearq"]
    app.rearq = rearq

    verbose = ctx.obj["verbose"]

    @app.on_event("startup")
    async def startup():
        await rearq.init()

    @app.on_event("shutdown")
    async def shutdown():
        await rearq.close()

    uvicorn.run(app=app, host=host, port=port, debug=verbose)


def main():
    sys.path.insert(0, ".")
    cli()


if __name__ == '__main__':
    main()



from fastapi import Depends, FastAPI, HTTPException
from starlette.status import HTTP_404_NOT_FOUND

from .. import ReArq
from ..job import Job
from .depends import get_rearq
from .schemas import AddJobIn

app = FastAPI()


@app.get("/job")
async def get_job(job_id: str, queue: str = 'default', rearq: ReArq = Depends(get_rearq)):
    job = Job(rearq.get_redis(), job_id, queue)
    return await job.info()


@app.get("/job/result")
async def get_job_result(job_id: str, queue: str = 'default', rearq: ReArq = Depends(get_rearq)):
    job = Job(rearq.get_redis(), job_id, queue)
    return await job.result_info()


@app.post("/job")
async def add_job(add_job_in: AddJobIn, rearq: ReArq = Depends(get_rearq)):
    task = rearq.get_task_map().get(add_job_in.task)
    if not task:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail='No such task')
    job = await task.delay(args=add_job_in.args, kwargs=add_job_in.kwargs, job_id=add_job_in.job_id)
    return await job.info()


@app.delete("/job")
async def cancel_job(job_id: str, rearq: ReArq = Depends(get_rearq)):
    ret = await rearq.cancel(job_id)
    return ret



import datetime
from typing import Any, List, Optional, Union

from pydantic import BaseModel


class AddJobIn(BaseModel):
    task: str
    args: Optional[List[Any]]
    kwargs: Optional[dict]
    job_id: Optional[str]
    countdown: Optional[Union[float, datetime.timedelta]]
    eta: Optional[datetime.datetime]
    expires: Optional[Union[float, datetime.datetime]]
    job_retry: int = 0



job_key_prefix = "rearq:job:"
in_progress_key_prefix = "rearq:in-progress:"
result_key_prefix = "rearq:result:"
retry_key_prefix = "rearq:retry:"
queue_key_prefix = "rearq:queue:"

health_check_key_suffix = "rearq:health-check"

default_queue: str = f"{queue_key_prefix}default"
delay_queue = f"{queue_key_prefix}delay"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rearq', 'rearq.api']

package_data = \
{'': ['*']}

install_requires = \
['aioredis', 'click', 'crontab', 'fastapi', 'pydantic', 'uvicorn']

entry_points = \
{'console_scripts': ['rearq = rearq.cli:main']}

setup_kwargs = {
    'name': 'rearq',
    'version': '0.2.0',
    'description': 'Rewrite arq and make improvement.',
    'long_description': '# Rearq\n\n![pypi](https://img.shields.io/pypi/v/rearq.svg?style=flat)\n\n## Introduction\n\nRearq is a distributed task queue with asyncio and redis, which rewrite from [arq](https://github.com/samuelcolvin/arq)\nand make improvement.\n\n## Install\n\nJust install from pypi:\n\n```shell\n> pip install rearq\n```\n\n## Quick Start\n\n### Task Definition\n\n```python\n# main.py\nrearq = ReArq()\n\n\n@rearq.on_shutdown\nasync def on_shutdown():\n    # you can do some clean work here like close db and so on...\n    print("shutdown")\n\n\n@rearq.on_startup\nasync def on_startup():\n    # you should do some initialization work here, such tortoise-orm init and so on...\n    print("startup")\n\n\n@rearq.task(queue="myqueue")\nasync def add(self, a, b):\n    return a + b\n\n\n@rearq.task(cron="*/5 * * * * * *")  # run task per 5 seconds\nasync def timer(self):\n    return "timer"\n```\n\n### Run rearq worker\n\n```shell\n> rearq main:rearq worker -q myqueue\n```\n\n```log\n2020-06-04 15:37:02 - rearq.worker:92 - INFO - Start worker success with queue: myqueue\n2020-06-04 15:37:02 - rearq.worker:84 - INFO - redis_version=6.0.1 mem_usage=1.47M clients_connected=25 db_keys=5\n```\n\n### Run rearq timing worker\n\nIf you have timeing task, run another command also:\n\n```shell\n> rearq main:rearq worker -t\n```\n\n```log\n2020-06-04 15:37:44 - rearq.worker:346 - INFO - Start timer worker success with queue: myqueue\n2020-06-04 15:37:44 - rearq.worker:84 - INFO - redis_version=6.0.1 mem_usage=1.47M clients_connected=25 db_keys=5\n```\n\n### Integration in FastAPI\n\n```python\napp = FastAPI()\n\n\n@app.on_event("startup")\nasync def startup() -> None:\n    await rearq.init()\n\n\n@app.on_event("shutdown")\nasync def shutdown() -> None:\n    await rearq.close()\n\n\n# then run task in view\n@app.get("/test")\nasync def test():\n    job = await add.delay(args=(1, 2))\n    return job.info()\n```\n\n## Rest API\n\nThere are several apis to control rearq.\n\n### Start server\n\n```shell\n> rearq main:rearq server\nUsage: rearq server [OPTIONS]\n\n  Start rest api server.\n\nOptions:\n  --host TEXT         Listen host.  [default: 0.0.0.0]\n  -p, --port INTEGER  Listen port.  [default: 8080]\n  -h, --help          Show this message and exit..\n```\n\n### API docs\n\nAfter server run, you can visit [https://127.0.0.1:8080/docs](https://127.0.0.1:8080/docs) to get all apis.\n\n### GET `/job`\n\nGet job information.\n\n### POST `/job`\n\nAdd a job for a task.\n\n### DELETE `/job`\n\nCancel a delay task.\n\n### GET `/job/result`\n\nGet job result.\n\n## Documentation\n\n> Writing...\n\n## ThanksTo\n\n- [arq](https://github.com/samuelcolvin/arq), Fast job queuing and RPC in python with asyncio and redis.\n\n## License\n\nThis project is licensed under the [MIT](https://github.com/long2ice/rearq/blob/master/LICENSE) License.\n',
    'author': 'long2ice',
    'author_email': 'long2ice@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/long2ice/rearq.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
